#include <CTPP2Types.h>

#include <map>
#include <string>
#include <vector>

template<typename T>class SymbolTable
{
public:
	SymbolTable();

	INT_32 GetSymbolId(CCHAR_P szSymbolName, const UINT_32 & iSymbolNamelength) const;

	INT_32 AddSymbol(CCHAR_P szSymbolName, const UINT_32 & iSymbolNamelength);

	INT_32 MarkScope();

	INT_32 UnmarkScope();

//	INT_32 RemoveSymbol(CCHAR_P szSymbolName, const UINT_32 & iSymbolNamelength);
//
//	INT_32 RemoveSymbol(const UINT_32 & iSymbolId);

	~SymbolTable() throw()
	{
		fprintf(stderr, "iLastVarId = %d\n", iLastVarId);
	}

private:
	/** Last variable id                   */
	INT_32     iLastVarId;
	/** Symbol table                       */
	std::map<std::string, std::vector<INT_32> > mSymbolTable;

	/** Last variable id                   */
	INT_32     iCurrentScope;
	/** Last variable id in current scope  */
	struct ScopeVars
	{
		std::vector<std::string>   varlist;
		INT_32                     last_var_id;
		ScopeVars(const std::vector<std::string> & vVarList, const UINT_32 & iLastVarId): varlist(vVarList), last_var_id(iLastVarId) { ;; }

	};
	std::vector<ScopeVars>   vScopeLastVarList;
};

//
//
//
template<typename T> SymbolTable<T>::SymbolTable():iLastVarId(0), iCurrentScope(0)
{
	vScopeLastVarList.push_back(ScopeVars(std::vector<std::string>(), 0));
}

//
//
//
template<typename T> INT_32 SymbolTable<T>::GetSymbolId(CCHAR_P szSymbolName, const UINT_32 & iSymbolNamelength) const
{
	std::map<std::string, std::vector<INT_32> >::const_iterator itmSymbolTable = mSymbolTable.find(std::string(szSymbolName, iSymbolNamelength));
	if (itmSymbolTable == mSymbolTable.end()) { return -1; }

	INT_32 iPos = itmSymbolTable -> second.size() - 1;

	if (iPos < 0) { return -1; }

return itmSymbolTable -> second[iPos];
}

//
//
//
template<typename T> INT_32 SymbolTable<T>::AddSymbol(CCHAR_P szSymbolName, const UINT_32 & iSymbolNamelength)
{
	std::string sSymbolName(szSymbolName, iSymbolNamelength);

	std::map<std::string, std::vector<INT_32> >::const_iterator itmSymbolTable = mSymbolTable.find(sSymbolName);
	if (itmSymbolTable != mSymbolTable.end())
	{
		typename std::vector<ScopeVars>::const_iterator itLastVarScopeId = vScopeLastVarList.end();
		--itLastVarScopeId;

		INT_32 iSymTablePos = itmSymbolTable -> second.size() - 1;
		INT_32 iScopeTablePos = vScopeLastVarList.size() - 1;

		if (vScopeLastVarList[iScopeTablePos].last_var_id <= itmSymbolTable -> second[iSymTablePos]) { return -1; }
	}

	vScopeLastVarList[iCurrentScope].varlist.push_back(sSymbolName);

	mSymbolTable[sSymbolName].push_back(iLastVarId);

return iLastVarId++;
}

//
//
//
template<typename T> INT_32 SymbolTable<T>::MarkScope()
{
	++iCurrentScope;
	vScopeLastVarList.push_back(ScopeVars(std::vector<std::string>(), iLastVarId));

return 0;
}

//
//
//
template<typename T> INT_32 SymbolTable<T>::UnmarkScope()
{
	typename std::vector<ScopeVars>::iterator itvCurrentScope = vScopeLastVarList.end();
	--itvCurrentScope;

	std::vector<std::string>::const_iterator itvVarList = itvCurrentScope -> varlist.begin();

	iLastVarId = itvCurrentScope -> last_var_id;

	while (itvVarList != itvCurrentScope -> varlist.end())
	{
		std::vector<INT_32> & oTMP = mSymbolTable[*itvVarList];
		std::vector<INT_32>::iterator itVarId = oTMP.end();
		--itVarId;
		oTMP.erase(itVarId);

		++itvVarList;
	}

	vScopeLastVarList.erase(itvCurrentScope);
	--iCurrentScope;

return 0;
}

int main(void)
{
	SymbolTable<INT_32> oIntTable;

	INT_32 iID = oIntTable.AddSymbol("a", 1);
	fprintf(stderr, "a(0) = %d\n", iID);

	iID = oIntTable.AddSymbol("b", 1);
	fprintf(stderr, "b(1) = %d\n", iID);

	iID = oIntTable.AddSymbol("c", 1);
	fprintf(stderr, "c(2) = %d\n", iID);

	iID = oIntTable.AddSymbol("a", 1);
	fprintf(stderr, "a(-1) = %d\n", iID);

	iID = oIntTable.GetSymbolId("c", 1);
	fprintf(stderr, "c(2) = %d\n", iID);

	oIntTable.MarkScope();
	fprintf(stderr, "MarkScope\n");

	iID = oIntTable.AddSymbol("b", 1);
	fprintf(stderr, "    b(3) = %d\n", iID);

	oIntTable.MarkScope();
	fprintf(stderr, "    MarkScope\n");

	iID = oIntTable.AddSymbol("b", 1);
	fprintf(stderr, "        b(4) = %d\n", iID);

	iID = oIntTable.AddSymbol("z", 1);
	fprintf(stderr, "        z(4) = %d\n", iID);

	oIntTable.UnmarkScope();
	fprintf(stderr, "    UnmarkScope\n");

	iID = oIntTable.AddSymbol("c", 1);
	fprintf(stderr, "    c(4) = %d\n", iID);

	iID = oIntTable.AddSymbol("d", 1);
	fprintf(stderr, "    d(5) = %d\n", iID);

	iID = oIntTable.GetSymbolId("c", 1);
	fprintf(stderr, "    c(4) = %d\n", iID);

	oIntTable.UnmarkScope();
	fprintf(stderr, "UnmarkScope\n");

	iID = oIntTable.GetSymbolId("a", 1);
	fprintf(stderr, "a(0) = %d\n", iID);

	iID = oIntTable.GetSymbolId("b", 1);
	fprintf(stderr, "b(1) = %d\n", iID);

	iID = oIntTable.GetSymbolId("c", 1);
	fprintf(stderr, "c(2) = %d\n", iID);

	iID = oIntTable.GetSymbolId("d", 1);
	fprintf(stderr, "d(-1) = %d\n", iID);

}
